(** * Introduction: tactics for the step-by-step derivation of programs from specifications *) 

(** The purpose here is to have an interactive derivation/refinement
   which remains readable and for which the proof of
   equivalence/refinment beetween each step is certified by Coq.  This
   proof should be as automatic as possible so that the user can focus
   on the derivation itself. *)

Require Import Strings.String.
Require Export Setoid.
Require Export Relation_Definitions.
Require Import Classes.RelationClasses.

(* begin hide *)

Ltac whileevar a eve evf neve nevf  := 
  match a with 
      context e [?a]=>is_evar a; 
                      let A:=fresh "EVAR" in 
                      eve A a; 
                        let x := evf A a in 
                        let  y := context e[x] in
                        whileevar y  ltac:eve ltac:evf ltac:neve ltac:nevf
    | _  => let x:=nevf a in  neve a
  end.

Ltac id a := a.

Ltac fst a b := a.

Ltac introEvar A a  := 
  match type of a with 
      ?Ta =>   evar (A: Ta)
  end.

Ltac print a := idtac a.

Ltac assert a := assert a.

Ltac name_evar t :=  whileevar t ltac:introEvar ltac:fst ltac:assert ltac:id .

Ltac name_evar_goal := match goal with |-?G => name_evar G end.

Inductive state : Prop :=
  RHS : state
| LHS : state
| GOAL : state
| ADHOC : forall s:string, state.  

Inductive memo (s: state) : Prop :=
  mem  :  memo s.

Tactic Notation "memorize" constr(s) := 
  pose ( mem _ : memo s).

Tactic Notation "setRHS" := 
  first [(match goal with 
            | [ H  : memo RHS |- _]=> idtac
            | [ H  : memo LHS |- _]=> clear H; memorize RHS
            | [ H  : memo GOAL |- _]=> clear H; memorize RHS
            | [ H  : memo (ADHOC ?s) |- _]=> memorize RHS
          end) | 
         (memorize RHS)].

Tactic Notation "setLHS" := 
  first [(match goal with 
            | [ H  : memo RHS |- _]=> clear H ; memorize LHS
            | [ H  : memo LHS |- _]=> idtac 
            | [ H  : memo GOAL |- _]=> clear H ; memorize LHS
            | [ H  : memo (ADHOC ?s) |- _]=> memorize LHS
  end) | 
  (memorize LHS)].

Tactic Notation "setGOAL" := 
  first [(match goal with 
    | [ H  : memo RHS |- _]=> clear H ; memorize GOAL
    | [ H  : memo LHS |- _]=> clear H ; memorize GOAL
    | [ H  : memo GOAL |- _]=> idtac 
    | [ H  : memo (ADHOC ?s) |- _]=> memorize GOAL
  end) | (memorize GOAL)].

Tactic Notation (at level 0) "if_RHS_LHS_BOTH_NOTH" 
       tactic(trhs) tactic(tlhs) tactic(tboth) tactic(tnoth) :=
  (lazymatch goal with
    | [ H : memo RHS |- _]=> trhs
    | [ H : memo LHS |- _]=>  tlhs
    | [ H : memo GOAL |- _]=> tboth
    | _ => tnoth
  end).

Tactic Notation  (at level 0) "if_RHS_LHS_BOTH" tactic(trhs) tactic(tlhs) tactic(tboth)
:= if_RHS_LHS_BOTH_NOTH  (trhs) (tlhs) (tboth) (fail "no side indication").

Tactic Notation "is_equivalence" constr(R) :=
  let H:= fresh in assert (Equivalence R) by (typeclasses eauto); clear H.

(* end hide *)

(** * Common tactics *)

(** ** Selection of the side to rewrite/refine in the equation *)

(** Right side.  *)

Tactic Notation  (at level 1)  "RHS" tactic(t) :=
  (if_RHS_LHS_BOTH_NOTH
    (idtac) 
    (idtac;
      match goal with 
        | [|- (?R ?lhs  ?rhs)] =>  subst rhs
      end
    ) 
    (idtac)
    (idtac)
  ) ;
  setRHS;
  match goal with 
    | [|-?R ?lhs ?rhs] =>  is_equivalence R;let LHS := fresh "LHS" in set (LHS:=lhs);
      assert(R LHS   rhs);[|assumption]
  end
  ;
  t.

(** Left side.  *)
Tactic Notation (at level 1) "LHS" tactic(t) :=
  (if_RHS_LHS_BOTH_NOTH
  (idtac;
    match goal with 
      | [|-?R ?lhs ?rhs] =>  subst lhs
    end
  ) 
  (idtac
  ) 
  (idtac)
  (idtac)
  );
  setLHS;
  match goal with 
    | [|-?R ?lhs ?rhs] =>  is_equivalence R; let RHS := fresh "RHS" in set (RHS:=rhs);
      assert(R lhs   RHS);[|assumption]
  end;
  t
.

(** Both sides. *)
Tactic Notation  (at level 1)  "GOAL" tactic(t) :=
  (if_RHS_LHS_BOTH_NOTH
  (idtac;
    match goal with 
      | [|-?R ?lhs ?rhs] =>  subst lhs
    end
  ) 
  (idtac;
    match goal with 
      | [|-?R ?lhs ?rhs] =>  subst rhs
    end
  ) 
  (idtac)
  (idtac)
  );
  setGOAL;
  t
  .

(* begin hide *)

(* debbuging tactics *)
Ltac print_goal :=
  match goal with
    [|-?G]=> idtac G
end.

(* Elimination of existentially quantified variable. *)
Tactic Notation "elim_ex" constr(e):=
  repeat (
    match (type of e) with 
      | exists x,_ => econstructor 
      | sig _ => econstructor 
      | _ => fail
    end
    )
.
Tactic Notation "eexistss" := 
  repeat( match goal with 
            | [|- exists A :?T,_] => 
              eexists
            | [|-sig _] => eexists
          end
 ).

(* end hide *)

(** ** Starting/ending a derivation/refinement session *)

(** Starts an interactive derivation session. *)
Ltac Begin := 
  intros;
  match goal with 
    | [|-?R ?lhs ?rhs] => is_equivalence R; idtac
    | [|- exists A ,_] => eexists;Begin
    | [|- sig _] => eexists;Begin
    | _ => eexists;intros
    | _ => fail "Error: no equivalence relation"
  end.

(** Starts an interactive derivation session with named existential variables. *)
Ltac BeginEvar := name_evar_goal.

(** Ends an interative derivation session with named existential variables. *)
Ltac endEvar :=     
  repeat  
    match reverse goal with
      | [var : ?A |- _] => 
        (lazymatch type of var with memo (ADHOC "stop") => fail  
        | ?a => idtac end);subst var
    end;
  match  goal with
    | [var : ?A |- _] => apply var
  end.

(** Starts an interactive derivation session in debugging mode. *)
Ltac BeginDebug := 
  intros;
  match goal with 
    | [|-?R ?LHS ?rhs] =>  is_equivalence R;idtac "Goal";print_goal
    | [|- exists A :?T,?B] =>  
      idtac "Existential goal ";print_goal; eexists; BeginDebug
  end.

(** Ends an interactive derivation session. *)
Tactic Notation "□" := 
  GOAL (reflexivity || let Stop := fresh "Stop" in 
  (pose (Stop := mem (ADHOC "stop"));
    repeat  
      match reverse goal with
        | [var : ?A |- _] => (lazymatch type of var with memo (ADHOC "stop") => fail  | ?a => idtac end);subst var
  end
  ; clear Stop); reflexivity).

(** **  Current goal verification *)

(** Checks that [e] is the [LHS]. *)
Tactic Notation "check" constr(e) "is" "LHS" :=
match goal with 
  [|- ?R e ?RHS] =>  is_equivalence R
  | _ => fail "Check failed: " e " is not the LHS"
end.

(** Checks that [e] is the [RHS]. *)
Tactic Notation "check" constr(e) "is" "RHS" := 
  match goal with 
      [|- ?R ?LHS e] =>  is_equivalence R
    | _ => fail "Check failed: " e " is not the RHS"
  end.

(** Checks that [e] is the [GOAL]. *)
Tactic Notation "check" constr(e) "is" "GOAL" :=
  match goal with 
      [|-e] =>  idtac
    | _ => fail "Check failed : " e " is not the GOAL"
  end.

(** * Program transformation tactics (equivalence) *)

(** This tactic insures equivalence between initial program and
   transformed program. *)

Tactic Notation  (at level 2) "=" "{"tactic(t1) "}" constr(e2) :=
  if_RHS_LHS_BOTH
    (match goal with
       | [|-?R ?lhs ?rhs] =>
         is_equivalence R;
       ((let h := fresh "rewriting" in 
         assert(h : R rhs e2) by (t1;try reflexivity) ; 
         rewrite h)
          || (change (rhs) with (e2))
       )
     end ||
         (t1; 
          (match goal with
             | [|-?R ?lhs e2] => is_equivalence R
           end) )
         || fail "Cannot derive to " e2 " from RHS by tactic " t1
    )
    (idtac;
     ((match goal with
           [|-?R ?lhs ?rhs] =>
           is_equivalence R;
           ((let h := fresh "rewriting" in 
             assert(h : R lhs e2) by (t1; try reflexivity) ;
             (rewrite h || change (lhs) with (e2)) )
              || (change (lhs) with (e2))
           )
       end) 
        || (t1; (match goal with
                   | [|-?R e2 ?rhs] => is_equivalence R
                 end)
           )
        || fail "Cannot derive to " e2 " from LHS by tactic " t1
     )
    )
    ( 
      (t1; check e2 is GOAL)
        ||
        (idtac;     
         match goal with
             [|-?R ?lhs ?rhs] =>
             is_equivalence R;
             match e2 with
               | ?R ?lhs_e2  ?rhs_e2 =>
                 is_equivalence R;
               let h := fresh "rewriting" in 
               assert(h : R lhs lhs_e2) by (t1;try reflexivity) ;
               rewrite h;
               let h2 := fresh "rewriting" in 
               assert(h : R rhs rhs_e2) by (t1;try reflexivity) ;
               rewrite h2
             end
         end
        )
        || fail "Cannot derive to " e2 " from GOAL by tactic " t1
    ).

Tactic Notation (at level 2 )   "=" "{" tactic(t1) "}d" constr(e2) :=
 info(  = { t1 } e2).

(** Starting form a goal with a [ LHS = RHS ] form, this tactic adds [ LHS = e2 ]
   in hypothesis if it can be proved by one rewritting step at place
   given by the function [f] using the equality [e2] then [LHS] is rewritten
   into [e2] in the goal. *)

(** For example, if we want to rewrite a goal [ x+x+x+x+x ] as [
2x+x+2x ] using [ H: forall x, x+x=2x], we would write :
<<
    LHS
    ={ pat (fun l=> l+x+l) (H x) }
    2*x+x+2*x  
>>
*)

Tactic Notation "pat" constr(f) constr(e1)  :=
  refine (eq_ind_r f _ e1) || fail "Cannot derive using pattern" f.

(** Derivations by definition of a given term. *)
Tactic Notation "by" "def" reference(e1) :=
  unfold e1 in *.

(** * Program transformation tactics (refinement) *)

(** To be included. *)