Require Import Coq.Lists.List.
Import ListNotations.
Require Import ProgramCalculation.theories.Algebra.Algebra.
Require Import ProgramCalculation.library.Notations.
Require Import ProgramCalculation.theories.Lists.Notations.
Require Import ProgramCalculation.library.Tactics.
Generalizable All Variables.

(** * [reduce] seen as defined on join-lists *)

(** [reduce] is defined using the usual cons-list in Coq. However in
    Richard Bird's "Lectures on Constructive Functional Programming",
    Oxford University Computing Laboratory, Technical Monograph
    PRG-69, 1988, [reduce] is defined on join-lists by three
    equations. This module provides a lemma proving that each
    equations holds for one definition of [reduce]. *)

Section Reduce_equations.

    Hypothesis A : Type.
    Hypothesis op : A -> A -> A.
    Hypothesis e : A.
    Infix "⊕" := op (at level 20).
    Notation "(⊕)" := (fun x y=>x⊕y).
    Notation "id⊕" := e.
    Context `{m:Monoid A op id⊕}.

    (** ** [ (⊕) / [] = id⊕ ]*)
    Lemma reduce_nil: (⊕) / [] = id⊕.
    Proof. auto. Qed.

    (** ** [ (⊕) / [a] = a ] *)
    Lemma reduce_one_point `(a:A): 
      (⊕) / [a] = a.
    Proof. unfold reduce. simpl. apply left_unit. Qed.

    Lemma fold_left_assoc:
      forall xs a b, fold_left (⊕) xs (a ⊕ b) = a ⊕ (fold_left (⊕) xs b) .
    Begin.
      generalize a b; clear a b.
      induction xs as [ | x xs induction_hypothesis]; intros a b.
      - trivial.
      - check ( fold_left (⊕) (x :: xs) (a ⊕ b) ) is LHS. 
        LHS
        = { simpl; rewrite induction_hypothesis }
        ( (a⊕b)⊕(fold_left (⊕) xs x) ).
        RHS 
        = { simpl; rewrite induction_hypothesis } 
        ( a⊕(b⊕(fold_left (⊕) xs x) ) ).
        = { rewrite associative }
        ( (a⊕b)⊕(fold_left (⊕) xs x) ).
        □.
    Qed.

    Lemma reduce_cons `(x:A)(xs:list A):
      (⊕) / (x::xs) = x ⊕ ((⊕) / xs).
    Begin.
      LHS 
      = { by def reduce }
      ( fold_left (⊕) xs (id⊕ ⊕ x) ).
      = { rewrite left_unit, right_unit } 
      ( fold_left (⊕) xs (x ⊕ id⊕) ).
      = { rewrite fold_left_assoc }
      ( x ⊕ (fold_left (⊕) xs id⊕ ) ).
      = { by def reduce}
      ( x ⊕ ((⊕) / xs) ).
      □.
    Qed.

    (** ** [ (⊕) / (x ++ y) = ( (⊕) / x ) ⊕ ( (⊕) / y ) ] *)
    Lemma reduce_join `(x : list A)(y : list A):
      (⊕) / (x ++ y) = ( (⊕)/x ) ⊕ ( (⊕)/y ).
    Begin. 
      induction x as [ | a x induction_hypothesis ].
      - check ( ((⊕) /) ([] ++ y) = 
                (((⊕) /) []) ⊕ (((⊕) /) y) ) is GOAL.
        LHS 
        = { rewrite left_unit }
        ( (⊕) / y ).
        RHS 
        = { rewrite left_unit }
        ( (⊕) / y ).
        □.
      - check ( ((⊕) /) ((a :: x) ++ y) = 
                (((⊕) /) (a :: x)) ⊕ (((⊕) /) y) ) is GOAL.
        LHS
        = { rewrite app_comm_cons }
        ( ((⊕) /) (a :: x ++ y) ).
        = { rewrite reduce_cons } 
        ( a ⊕ ( (⊕) / (x ++ y) ) ).
        = { rewrite induction_hypothesis }
        ( a ⊕ ( ( (⊕)/x ) ⊕ ( (⊕)/y) ) ).
        RHS
        = { rewrite reduce_cons }
        ( ((a ⊕ ((⊕) / x)) ⊕ (((⊕) /) y) ) ).
        = { rewrite associative } 
        ( a ⊕ ( ( (⊕)/x ) ⊕ ( (⊕)/y) ) ).
        □.
    Qed.

End Reduce_equations.